unit enhlv;


interface

{$IFDEF VER90}
{$DEFINE GL_D2}
{$ENDIF}

{$IFDEF VER100}
{$DEFINE GL_D3}
{$ENDIF}

{$IFDEF VER120}
{$DEFINE GL_D3}
{$DEFINE GL_D4}
{$ENDIF}

{$IFDEF VER130}
{$DEFINE GLINCLUDEBUGFIX}
{$ENDIF}

{$IFDEF VER140}
{$DEFINE GLINCLUDEBUGFIX}
{$ENDIF}

uses comctrls, messages, classes, commctrl, sysutils, {$IFDEF GLINCLUDEBUGFIX}imglist, {$ENDIF}windows;

type
   TEnumToolTipIcon = (ttiNoIcon, ttiInfo, ttiWarning, ttiError);
   TEnumToolTipOptions = (ttoInstantShow, ttoLongStay, ttoBelowItem, ttoOffSet, ttoAdjustForScreenBoundary);
   TTooltipOptions = set of TEnumToolTipOptions;
{$IFDEF GL_D2}
   TOnToolTipEvent = procedure(sender: Tobject; item: TListItem; logicalCol: integer; var toolTipText: string; var toolTipTitle: string; var toolTipTitleIcon: TEnumToolTipIcon; var options : TTooltipOptions) of object;
{$ELSE}
   TOnToolTipEvent = procedure(sender: Tobject; item: TListItem; logicalCol: integer; var toolTipText: widestring; var toolTipTitle: widestring; var toolTipTitleIcon: TEnumToolTipIcon; var options : TTooltipOptions) of object;
{$ENDIF}
   TOnShowToolTipEvent = procedure(Sender: TObject; item: TListItem; logicalCol: integer; const fieldBound, oldTooltipBound: TRECT; var AdjustedTooltipLeft, AdjustedTooltipTop: integer) of object;

{
   Freeware copyright Greg Lorriman 1999-2002
   http://www.lorriman.com greg@lorriman.com
   Version 1.0

   Licensed for use by anyone.

   #### WARNING #### See NOTES if upgrading from previous versions of this component.

   Adds tooltips to listview for column text for Delphi 2/3/4/5/6

   If you have a genuine need for the ability to change font and color please email me.

   Set ShowTooltips to true. Adjust ToolTipOptions and ToolTipWidth. Use
   OnToolTip event for custom tooltip text. Ordinarily tooltip only displays
   when column text is truncated, but custom tooltip text forces display of
   tooltip. Use OnShowToolTip to change location of tooltip.
   You can use HandleToolTip for further manipulation. Later versions of
   comctrl32 support tooltip title and icon; test SupportsToolTipTitle in
   OnToolTip event and adjust tooltTipTitle and toolTipTitleIcon parameters.

   NOTES: Functionality requires update to Comctrl32.dll (IE4+).
   vsIcon view has a fixed tip width : ToolTipWidth has no effect.
   ttoLongStay=32 seconds is a win32 limit.
   D3+ supports unicode via widestring for ToolTipText and ToolTipTitle; code
   adjustment will be needed if upgrading project from D2.
   Previous versions of this component will have an incompatible OnToolTip
   event; adjust your event handlers to refelct the new and reordered parameters.
   /Incorporates D5 virtual listview memory leak fix (see source). See http://community.borland.com for
   XP listview fix (without which the listview will not work in XP).

   See http://msdn.microsoft.com/library/sdkdoc/shellcc/CommCtls/ListView/Using.htm
   and http://msdn.microsoft.com/library/sdkdoc/shellcc/CommCtls/ToolTip/ToolTip.htm

   for implementation info. Remember to "sync toc".

   DISCLAIMER :

   THIS SOFTWARE AND THE ACCOMPANYING FILES ARE SOLD "AS IS" AND WITHOUT WARRANTIES AS TO PERFORMANCE OF MERCHANTABILITY OR ANY OTHER WARRANTIES WHETHER EXPRESSED OR IMPLIED. Because of the various hardware and software environments into which F1Help may be put, NO WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE IS
   OFFERED. Good data processing procedure dictates that any program be thoroughly tested with non-critical data before relying on it.The user must assume the entire risk of using the program.  ANY LIABILITY OF THE SELLER WILL BE LIMITED EXCLUSIVELY TO PRODUCT REPLACEMENT OR REFUND OF PURCHASE PRICE.




}
   TgfglListView = class(TListview)
   private

      FToolTipOptions: TToolTipOptions;
      FTransientToolTipOptions : TToolTipOptions;
{$IFDEF GL_D2}
      FToolTipText: string;
{$ELSE}
      FToolTipText: widestring;
{$ENDIF}
      FToolTipwidth: integer;
      FOnToolTip: TOnToolTipEvent;
      FHelp: TStrings;
      FOnShowToolTip: TOnShowToolTipEvent;
      FSupportsToolTiptitle: boolean;

      procedure SetVersion(const Value: string);
      function GetVersion: string;
      function getOtherstuff: string;
      procedure setOtherStuff(const Value: string);
      procedure setHelp(const Value: TStrings);
      procedure adjustForScreenBoundaries(var x, y: integer;
         position: TRECT);
      function getHandleToolTip: HWND;

   protected
{$IFDEF GLINCLUDEBUGFIX}
   //This is for the incorporation of a critical memory leak bug fix. See Sterkenburg's site
   //This bug fix does not solve the XP problem for which you need to copy the comctrls unit
   //into your project folder, modify it to fix the problem (as documented on http://community.borland.com)
   //and then compile. Using the copy unit method precludes the use of packages.
   //The XP bug cannot be fixed any other way.
      function CreateListItem: TListItem; override;
{$ENDIF}

      procedure WndProc(var Message: TMessage); override;

      function getToolTipWidth: integer; virtual;
      procedure setToolTipWidth(const Value: integer); virtual;
      function GetToolTips: boolean; virtual;
      procedure SetToolTipOptions(const Value: TToolTipOptions); virtual;
      procedure SetToolTips(const Value: boolean); virtual;
{$IFDEF GL_D2}
      procedure doOnToolTip(item: TListItem; logicalCol: integer; var toolTipText: string; var toolTipTitle: string; var toolTipTitleIcon: TEnumToolTipIcon; var options : TTooltipOptions); virtual;
{$ELSE}
      procedure doOnToolTip(item: TListItem; logicalCol: integer; var toolTipText: widestring; var toolTipTitle: widestring; var toolTipTitleIcon: TEnumToolTipIcon; var options : TTooltipOptions); virtual;
{$ENDIF}
      procedure doShowToolTip(Sender: TObject; item: TListItem; logicalCol: integer; const fieldBound, tooltipBound: TRECT; var tooltipLeft, tooltipTop: integer); virtual;

   public

      constructor Create(AOwner: Tcomponent); override;
      destructor Destroy; override;

      {
      Tests for whether version of Comctrl32.dll supports tooltip titles.
      Use in OnToolTip event if the ToolTip text must be different depending on
      whether a title will be displayed.
      }
      function SupportsTooltipTitle: boolean;
      {
      Handle to the tooltip control. Use for further manual manipulation.
      }
      property HandleToolTip: HWND read getHandleToolTip;

   published
      {Enables or disables tooltips}
      property ShowToolTips: boolean read GetToolTips write SetToolTips;
      {
         -ttoInstantShow = tooltip instantly displays

         -ttoLongStay = tooltip hangs around for a 32 seconds

         -ttoBelowItem = Tooltip displays below the item's column text

         -ttoOffSet = tooltip displays slightly to the right

         -ttoAdjustForScreenBoundary = tooltip is adjusted if it would otherwise fall off the edge of the screen

         These options can be modified within the OnToolTip event, but for specific modifications
         use the "Options" parameter of the event to avoid affecting the overall options.
      }
      property ToolTipOptions: TToolTipOptions read FToolTipOptions write SetToolTipOptions;
{
   ToolTipWidth<>-1 results in multi-line tip, otherwise no width limit.
   Force multi-line with #13#10 combination (ToolTipWidth>-1)
}
      property ToolTipWidth: integer read getToolTipWidth write setToolTipWidth;

      property Version: string read GetVersion write SetVersion stored false;
      property OtherStuff: string read getOtherstuff write setOtherStuff stored false;
      property Help: TStrings read FHelp write setHelp stored False;
      {
      You can modify most of this event's parameters to customise display of tooltip
      for specific item columns.


      item : TListItem -  The item for which the tooltip is to be displayed. Not modifiable.

      logicalCol : integer - The column counting from 0 (0=caption, 1=subItems[0]). Not modifiable.

      toolTipText : widestring (or string in D2) - The text that will appear. If blank the tooltip will not appear unless
         you modify it. Accepts unicode in D3+.

      toolTipTitle : widestring (or string in D2) -  The title text if supported (see SupportToolTipTitle). Defaults to
         blank (no title). Modify to display title. Accepts unicode in D3+.

      toolTipTitleIcon: TEnumToolTipIcon -  Accompanies title text. One of ttiNoIcon, ttiInfo, ttiWarning, ttiError. Defaults ot ttiNoIcon.

      options : TTooltipOptions - Adjusting this parameter avoids affecting overall options.

      }
      property OnToolTip: TOnToolTipEvent read FOnToolTip write FOnToolTip;
      {
      For  modifying the x,y position of the tooltip. Uses screen coords.

            -fieldBound = bounding rectabgle of column item

            -oldTooltipBound = original (unadjusted) bounding rectangle of tooltip

            -var newTooltipLeft = adjusted x position according to options

            -var newTooltipTop = adjusted y position according to options

            You can modify the last two parameters.

      For further manipulation see HandleToolTip
      }
      property OnShowToolTip: TOnShowToolTipEvent read FOnShowToolTip write FOnShowToolTip;

   end;

procedure Register;

implementation

uses controls;

function GetBuildInfo(filename: string; var V1, V2, V3, V4: Integer): boolean;
   { by Steve Schafer }
var
   VerInfoSize: DWORD;
   VerInfo: Pointer;
   VerValueSize: DWORD;
   VerValue: PVSFixedFileInfo;
   Dummy: DWORD;
begin
   result := false;
   try
      VerInfoSize := GetFileVersionInfoSize(PChar(filename), Dummy);
      if VerInfoSize = 0 then EXIT;
      GetMem(VerInfo, VerInfoSize);
      try
         if not GetFileVersionInfo(PChar(filename), 0, VerInfoSize, VerInfo) then EXIT;
         if not VerQueryValue(VerInfo, '\', Pointer(VerValue), VerValueSize) then EXIT;
         with VerValue^ do begin
            V1 := dwFileVersionMS shr 16;
            V2 := dwFileVersionMS and $FFFF;
            V3 := dwFileVersionLS shr 16;
            V4 := dwFileVersionLS and $FFFF;
         end;
      finally
         FreeMem(VerInfo, VerInfoSize);
      end;
      result := true;
   except
   //for some reason this funciton occasionally excepted. I;ve adjusted it (better error checking)
   //which means this trap may be redundant.
   end;
end;

const

   LVM_FIRST = $1000;
   LVM_SETEXTENDEDLISTVIEWSTYLE = LVM_FIRST + 54;
   LVM_GETEXTENDEDLISTVIEWSTYLE = LVM_FIRST + 55;
   LVS_EX_INFOTIP = $00000400;
   LVM_GETTOOLTIPS = LVM_FIRST + 78;
   TTM_SETDELAYTIME = WM_USER + 3;
   TTDT_AUTOMATIC = 0;
   TTDT_AUTOPOP = 2;
   TTM_SETMAXTIPWIDTH = WM_USER + 24;
   TTN_FIRST = 0 - 520;
   TTN_NEEDTEXTW = TTN_FIRST - 10;
   TTN_GETDISPINFOW = TTN_NEEDTEXTW;
   TTM_GETMAXTIPWIDTH = WM_USER + 25;
   LVN_GETINFOTIPW = LVN_FIRST - 58;
   LVM_SUBITEMHITTEST = LVM_FIRST + 57;
   LVM_GETITEMRECT = LVM_FIRST + 14;
   TTM_SETTITLEA = WM_USER + 32;
   TTM_SETTITLEW = WM_USER + 33;

type
   tagNMTTDISPINFOW = packed record
      hdr: TNMHdr;
      lpszText: PWideChar;
      szText: array[0..79] of WideChar;
      hinst: HINST;
      uFlags: UINT;
      lParam: LPARAM;
   end;
   PNMTTDispInfoW = ^TNMTTDispInfoW;
   TNMTTDispInfoW = tagNMTTDISPINFOW;

type
   tagLVHITTESTINFO = packed record
      pt: TPoint;
      flags: UINT;
      iItem: Integer;
      iSubItem: Integer;
   end;
   TLVHitTestInfo = tagLVHITTESTINFO;

{ TGFGLListView }

constructor TgfglListView.create(AOwner: Tcomponent);
const helptext: string = '   Freeware copyright Greg Lorriman 1999-2002'#13#10+
'   http://www.lorriman.com greg@lorriman.com'#13#10+
'   See source for version number.'#13#10+
''#13#10+
'   Licensed for use by anyone.'#13#10+
''#13#10+
'   #### WARNING #### See NOTES if upgrading from previous versions of this component.'#13#10+
''#13#10+
'   Adds tooltips to listview for column text for Delphi 2/3/4/5/6'#13#10+
''#13#10+
'   If you have a genuine need for the ability to change font and color please email me.'#13#10+
''#13#10+
'   Set ShowTooltips to true. Adjust ToolTipOptions and ToolTipWidth. Use'#13#10+
'   OnToolTip event for custom tooltip text. Ordinarily tooltip only displays'#13#10+
'   when column text is truncated, but custom tooltip text forces display of'#13#10+
'   tooltip. Use OnShowToolTip to change location of tooltip.'#13#10+
'   You can use HandleToolTip for further manipulation. Later versions of'#13#10+
'   comctrl32 support tooltip title and icon; test SupportsToolTipTitle in'#13#10+
'   OnToolTip event and adjust tooltTipTitle and toolTipTitleIcon parameters.'#13#10+
''#13#10+
'   NOTES: Functionality requires update to Comctrl32.dll (IE4+).'#13#10+
'   vsIcon view has a fixed tip width : ToolTipWidth has no effect.'#13#10+
'   ttoLongStay=32 seconds is a win32 limit.'#13#10+
'   D3+ supports unicode via widestring for ToolTipText and ToolTipTitle; code'#13#10+
'   adjustment will be needed if upgrading project from D2.'#13#10+
'   Previous versions of this component will have an incompatible OnToolTip'#13#10+
'   event; adjust your event handlers to refelct the new and reordered parameters.'#13#10+
''#13#10+
'   See http://msdn.microsoft.com/library/sdkdoc/shellcc/CommCtls/ListView/Using.htm'#13#10+
'   and http://msdn.microsoft.com/library/sdkdoc/shellcc/CommCtls/ToolTip/ToolTip.htm'#13#10+
'   for implementation info. Remember to "sync toc".'#13#10+
''#13#10+
'   Disclaimer in readme.txt or contact Greg for text of disclaimer';


var
   a, b, c, d: integer;
begin
   inherited create(AOwner);
   if getBuildInfo('Comctl32.dll', a, b, c, d) then
      FSupportsToolTiptitle := (a >= 5) and (b >= 80);
   FToolTipWidth := -1;
   FHelp := TStringlist.create;
   if csDesigning in componentState then
      FHelp.text := helptext;
   FToolTipOptions := [ttoAdjustForScreenBoundary];
end;

destructor TgfglListView.destroy;
begin
   FHelp.free;
   inherited destroy;
end;

{$IFDEF GL_D2}

procedure TgfglListView.doOnToolTip(item: TListItem; logicalCol: integer; var toolTipText: string; var toolTipTitle: string; var toolTipTitleIcon: TEnumToolTipIcon; var options : TTooltipOptions);
begin
   if assigned(FonToolTip) then
      FonToolTip(self, item, logicalCol, toolTipText, toolTipTitle, toolTipTitleIcon,options);
end;
{$ELSE}

procedure TgfglListView.doOnToolTip(item: TListItem; logicalCol: integer; var toolTipText: widestring; var toolTipTitle: widestring; var toolTipTitleIcon: TEnumToolTipIcon; var options : TTooltipOptions);
begin
   if assigned(FonToolTip) then
      FonToolTip(self, item, logicalCol, toolTipText, toolTipTitle, toolTipTitleIcon,options);
end;
{$ENDIF}

function TGFGLListView.getOtherstuff: string;
begin
   result := 'http://www.lorriman.com';
end;

function TGFGLListView.GetToolTips: boolean;
begin
   result := (sendmessage(handle, LVM_GETEXTENDEDLISTVIEWSTYLE, 0, 0) and LVS_EX_INFOTIP) > 0;
end;

function TGFGLListView.getToolTipWidth: integer;
begin
   result := FTooltipwidth
end;

function TGFGLListView.GetVersion: string;
begin
   result := '1.0';
end;

procedure TgfglListView.setHelp(const Value: TStrings);
begin

end;

procedure TGFGLListView.setOtherStuff(const Value: string);
begin

end;

procedure TGFGLListView.SetToolTipOptions(const Value: TToolTipOptions);
var
   ttHandle: HWND;
   auto, autoPop: longint;
   val: TToolTipOptions;
begin
   val := value;
   if (ttoOffset in val) and not (ttoOffset in FToolTipOptions) then
      val := val - [ttoBelowItem];
   if (ttoBelowItem in val) and not (ttoBelowItem in FToolTipOptions) then
      val := val - [ttoOffset];
   FToolTipOptions := Val;
   if ttoInstantShow in FToolTipOptions then auto := 1 else auto := 500;
   if ttoLongStay in FToolTipOptions then autoPop := 32000 else autoPop := 3000;
   ttHandle := HandletoolTip;
   sendmessage(ttHandle, TTM_SETDELAYTIME, TTDT_AUTOMATIC, auto);
   sendmessage(ttHandle, TTM_SETDELAYTIME, TTDT_AUTOPOP, autopop);
end;

procedure TGFGLListView.SetToolTips(const Value: boolean);
var
   data: longint;
begin
   if value then data := LVS_EX_INFOTIP else data := 0;
   sendmessage(handle, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_INFOTIP, data)
end;

procedure TGFGLListView.setToolTipWidth(const Value: integer);
begin
   FTooltipWidth := value;
end;

procedure TGFGLListView.SetVersion(const Value: string);
begin

end;

procedure TGFGLListView.WndProc(var Message: TMessage);
var
   fHwnd: HWND;
   winplace: TWindowPlacement;
   hittest: TLVHitTestInfo;
   x, y: integer;
   item: TListItem;
   itemBound: TRect;
   cursorpos: TPoint;
{$IFDEF GL_D2}
   toolTipTitle: string;
{$ELSE}
   toolTipTitle: widestring;
{$ENDIF}
   toolTipTitleIcon: TEnumToolTipIcon;
   savedToolTipOptions : TTooltipOptions;
begin
   toolTipTitleIcon := ttiNoIcon;

   inherited wndproc(message);

   if csDesigning in componentstate then EXIT;

   if (message.msg = WM_NOTIFY) then begin
      fHwnd := PNMHdr(message.lparam)^.hwndFrom;
      case PNMHdr(message.lparam)^.code of
         TTN_GETDISPINFOW: begin
               SetToolTipOptions(FToolTipOptions); //in case of dynamic creation
               FTransientToolTipOptions:=FToolTipOptions;
               SendMessage(fHwnd, TTM_SETMAXTIPWIDTH, 0, FToolTipwidth);
               if assigned(FOnToolTip) then begin
                  getcursorpos(cursorpos);
                  cursorpos := ScreenToClient(cursorpos);
                  hittest.pt.x := cursorpos.x;
                  hittest.pt.y := cursorpos.y;
                  hittest.iItem := -1; //in case of unsupported hittest message
                  hittest.iSubitem := -1;
                  sendmessage(handle, LVM_SUBITEMHITTEST, 0, integer(@hittest));
                  if hittest.iItem >= 0 then begin
                     item := items[hittest.iItem];
{$IFDEF GL_D2}
                     if PNMTTDispInfoW(message.lparam)^.lpszText <> nil then
                        FToolTipText := widechartostring(PNMTTDispInfoW(message.lparam)^.lpszText)
                     else
                        FToolTipText := '';
                     doOnToolTip(item, hittest.iSubItem, FTooltipText, toolTipTitle, toolTipTitleIcon,FTransientToolTipOptions);
                     getmem(PNMTTDispInfoW(message.lparam)^.lpszText, sizeof(widechar) * length(FToolTipText) + 2);
                     stringToWideChar(FToolTipText, PNMTTDispInfoW(message.lparam)^.lpszText, length(FToolTipText) * 2);
                     if supportsTooltipTitle then
                        sendmessage(fHwnd, TTM_SETTITLEA, ord(toolTipTitleIcon), LPARAM(pchar(toolTipTitle)));
{$ELSE}
                     FToolTipText := widestring(PNMTTDispInfoW(message.lparam)^.lpszText);
                     doOnToolTip(item, hittest.iSubItem, FTooltipText, toolTipTitle, toolTipTitleIcon,FTransientToolTipOptions);
                     PNMTTDispInfoW(message.lparam)^.lpszText := pwidechar(FToolTipText);
                     if supportsTooltipTitle then
                        sendmessage(fHwnd, TTM_SETTITLEW, ord(toolTipTitleIcon), LPARAM(pchar(toolTipTitle)))
{$ENDIF}
                  end;
               end;
            end;
         TTN_SHOW: begin
               savedToolTipOptions:=FToolTipOptions;
               SetToolTipOptions(FTransientToolTipOptions); //in case of dynamic creation
               winplace.length := sizeof(winplace);
               getwindowplacement(fHwnd, @winplace);
               getcursorpos(cursorpos);
               cursorpos := ScreenToClient(cursorpos);
               hittest.pt.x := cursorpos.x;
               hittest.pt.y := cursorpos.y;
               hittest.iItem := -1; //in case of unsupported hittest message
               hittest.iSubitem := -1;
               sendmessage(handle, LVM_SUBITEMHITTEST, 0, integer(@hittest));
               if hittest.iItem >= 0 then begin
                  item := items[hittest.iItem];
                  itemBound.left := LVIR_LABEL;
                  sendmessage(handle, LVM_GETITEMRECT, hittest.iItem, integer(@itemBound));
                  y := winplace.rcNormalPosition.top;
                  x := winplace.rcNormalPosition.left;
                  if ttoOffset in FTransientToolTipOptions then begin
                     getcursorpos(cursorpos);
                     x := cursorpos.x + 20;
                     y := cursorpos.y + ((itemBound.bottom - itembound.top) div 2);
                  end else
                     if ttoBelowItem in FTransientToolTipOptions then
                        inc(y, itemBound.bottom - itembound.top);
                  if ttoAdjustForScreenBoundary in FTransientToolTipOptions then
                     adjustForScreenBoundaries(x, y, winplace.rcNormalPosition);
                  if assigned(FonShowToolTip) then
                     doShowToolTip(self, item, hittest.iSubItem, itemBound, winplace.rcNormalPosition, x, y);

                  setwindowpos(fHwnd, HWND_TOPMOST, x, y, 0, 0, SWP_NOSIZE or SWP_NOACTIVATE);
               end;
               FToolTipOptions:=savedToolTipOptions;
            end;
      end;
   end;
end;

procedure TGFGLListView.adjustForScreenBoundaries(var x, y: integer; position: TRECT);
var
   sW, sH: integer;
begin
   sW := getSystemMetrics(SM_CXSCREEN);
   sH := getSystemMetrics(SM_CYSCREEN);
   if x < 0 then x := 0;
   if y < 0 then y := 0;
   if (position.Right + (x - position.left)) > sW then
      x := x - (position.Right - sW) - (x - position.left);
   if (position.bottom + (y - position.top)) > sH then
      y := y - (position.bottom - sH) - (y - position.top);
end;

//##############Start of bug fix code##############

{$IFDEF GLINCLUDEBUGFIX}
{$HINTS off}
type
   TListItemCracker = class(TPersistent)
   private
      FOwner: TListItems;
      FSubItems: TStrings;
   end;
{$HINTS on}

   TReplacementListItem = class(TListItem)

   public
      constructor Create(AOwner: TListItems);

   end;

type
   TFixedSubItems = class(TStringList)
   private
      FOwner: TListItem;
      FImageIndices: TList;
      procedure SetColumnWidth(Index: Integer);
      procedure RefreshItem(Index: Integer);
      function GetImageIndex(Index: Integer): TImageIndex;
      procedure SetImageIndex(Index: Integer; const Value: TImageIndex);
   protected
      function GetHandle: HWND;
      function Add(const S: string): Integer; override;
      procedure Delete(Index: Integer); override;
      procedure Put(Index: Integer; const S: string); override;
      procedure SetUpdateState(Updating: Boolean); override;
   public
      constructor Create(AOwner: TListItem);
      destructor Destroy; override;

    //This is the fix
      procedure Clear; override;

      procedure Insert(Index: Integer; const S: string); override;
      property Handle: HWND read GetHandle;
      property Owner: TListItem read FOwner;
      property ImageIndex[Index: Integer]: TImageIndex read GetImageIndex write SetImageIndex;
   end;

constructor TFixedSubItems.Create(AOwner: TListItem);
begin
   inherited Create;
   FOwner := AOwner;
   FImageIndices := TList.Create;
end;

destructor TFixedSubItems.Destroy;
begin
   FImageIndices.Free;
   inherited;
end;

function TFixedSubItems.Add(const S: string): Integer;
begin
   Result := inherited Add(S);
   FImageIndices.Add(Pointer(-1));
   RefreshItem(Result + 1);
end;

procedure TFixedSubItems.Delete(Index: Integer);
begin
   inherited;
   FImageIndices.Delete(Index);
   Owner.Update;
end;

function TFixedSubItems.GetHandle: HWND;
begin
   Result := Owner.Owner.Handle;
end;

procedure TFixedSubItems.SetColumnWidth(Index: Integer);
var
   ListView: TCustomListView;
begin
   ListView := Owner.ListView;
   if TGFGLListView(ListView).ColumnsShowing and
      (TGFGLListView(ListView).Columns.Count > Index) and
      (ListView.Column[Index].WidthType = ColumnTextWidth) then
      TGFGLListView(ListView).UpdateColumn(Index);
end;

procedure TFixedSubItems.Insert(Index: Integer; const S: string);
var
   i: Integer;
begin
   inherited Insert(Index, S);
   FImageIndices.Insert(Index, Pointer(-1));
   for i := Index + 1 to Count do RefreshItem(i);
end;

procedure TFixedSubItems.Put(Index: Integer; const S: string);
begin
   inherited Put(Index, S);
   RefreshItem(Index + 1);
end;

procedure TFixedSubItems.RefreshItem(Index: Integer);
begin
   ListView_SetItemText(Handle, Owner.Index, Index, LPSTR_TEXTCALLBACK);
   SetColumnWidth(Index);
end;

type TListItemsCracker = class(TListItems)
   end;

procedure TFixedSubItems.SetUpdateState(Updating: Boolean);
begin
   TListItemsCracker(Owner.Owner).SetUpdateState(Updating);
end;

function TFixedSubItems.GetImageIndex(Index: Integer): TImageIndex;
begin
   Result := TImageIndex(FImageIndices[Index]);
end;

procedure TFixedSubItems.SetImageIndex(Index: Integer; const Value: TImageIndex);
begin
   FImageIndices[Index] := Pointer(Value);
end;

//This is the fix

procedure TFixedSubItems.Clear;
begin
   inherited Clear;
   FImageIndices.Clear;
end;
//#

constructor TReplacementListItem.Create(AOwner: TListItems);
begin
   inherited create(AOwner);
   SubItems.free;
   TListItemCracker(self).FSubItems := TFixedSubItems.create(self);
end;

function TgfglListView.CreateListItem: TListItem;
begin
   result := TReplacementListItem.Create(Items);
end;

{$ENDIF}
//##############End of bug fix code##############

procedure TgfglListView.doShowToolTip(Sender: TObject; item: TListItem; logicalCol: integer; const fieldBound, tooltipBound: TRECT; var tooltipLeft, tooltipTop: integer);
begin
   if assigned(FonShowToolTip) then
      FonShowToolTip(self, item, logicalCol, fieldBound, tooltipBound, tooltipLeft, tooltipTop);

end;

function TgfglListView.supportsTooltipTitle: boolean;
begin
   result := FSupportsToolTipTitle;
end;

function TgfglListView.getHandleToolTip: HWND;
begin
   result := sendmessage(handle, LVM_GETTOOLTIPS, 0, 0);
end;

procedure Register;
begin
   RegisterComponents({$IFDEF GL_D2} 'Win95'{$ELSE} 'Win32'{$ENDIF}, [TgfglListView]);
end;

end.

